/**
 * 
 */
package ch.psi.daq.test.queryrest.query;

import java.util.List;
import java.util.Random;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.TimeUnit;
import java.util.regex.Pattern;
import java.util.stream.Collectors;
import java.util.stream.LongStream;
import java.util.stream.Stream;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.google.common.collect.Lists;

import ch.psi.daq.cassandra.reader.CassandraReader;
import ch.psi.daq.cassandra.util.test.CassandraDataGen;
import ch.psi.daq.common.ordering.Ordering;
import ch.psi.daq.domain.DataEvent;
import ch.psi.daq.domain.cassandra.ChannelConfiguration;
import ch.psi.daq.domain.cassandra.ChannelEvent;
import ch.psi.daq.domain.cassandra.MetaPulseId;
import ch.psi.daq.domain.cassandra.query.PulseIdRangeQuery;
import ch.psi.daq.domain.cassandra.query.TimeRangeQuery;
import ch.psi.daq.domain.cassandra.querying.ChannelEventQuery;
import ch.psi.daq.domain.cassandra.querying.MetaChannelEvent;
import ch.psi.daq.domain.cassandra.querying.EventQuery;

public class DummyCassandraReader implements CassandraReader {
   private static final Logger LOGGER = LoggerFactory.getLogger(DummyCassandraReader.class);

   private static final int KEYSPACE = 1;
   private CassandraDataGen dataGen;
   private String[] channels;
   private Random random = new Random(0);

   /**
    * 
    */
   public DummyCassandraReader() {
      this.dataGen = new CassandraDataGen();
      this.channels = new String[] {
            "testChannel1",
            "testChannel2",
            "BoolScalar",
            "BoolWaveform",
            "Int8Scalar",
            "Int8Waveform",
            "UInt8Scalar",
            "UInt8Waveform",
            "Int16Scalar",
            "Int16Waveform",
            "UInt16Scalar",
            "UInt16Waveform",
            "Int32Scalar",
            "Int32Waveform",
            "UInt32Scalar",
            "UInt32Waveform",
            "Int64Scalar",
            "Int64Waveform",
            "UInt64Scalar",
            "UInt64Waveform",
            "Float32Scalar",
            "Float32Waveform",
            "Float64Scalar",
            "Float64Waveform",
            "StringScalar"};
   }

   /**
    * @{inheritDoc
    */
   @Override
   public Stream<String> getChannelStream(String regex) {
      Stream<String> channelStream = Stream.of(channels);
      if (regex != null) {
         Pattern pattern = Pattern.compile(regex.toLowerCase());
         channelStream = Stream.of(channels).filter(channel -> pattern.matcher(channel.toLowerCase()).find());
      }
      return channelStream;
   }

   /**
    * @{inheritDoc
    */
   @Override
   public Stream<? extends DataEvent> getEventStream(String channel, long startPulseId, long endPulseId,
         Ordering ordering, boolean aggregateValues, String... columns) {
      return getDummyEventStream(channel, startPulseId, endPulseId);
   }

   /**
    * @{inheritDoc
    */
   @Override
   public Stream<? extends DataEvent> getEventStream(String channel, long startMillis, long startNanos, long endMillis,
         long endNanos, Ordering ordering, boolean aggregateValues, String... columns) {
      return getDummyEventStream(channel, startMillis / 10, endMillis / 10);
   }

   /**
    * @{inheritDoc
    */
   @Override
   public Stream<ChannelEvent> getEventStream(EventQuery eventQuery, Stream<? extends ChannelEventQuery> queryProviders) {
      Stream<ChannelEvent> result = queryProviders.map(ceq -> {
         if (ceq instanceof MetaChannelEvent) {
            return getEvent((MetaChannelEvent) ceq);
         } else {
            throw new UnsupportedOperationException("This is not yet implemented!");
         }
      });

      return result;
   }

   /**
    * @{inheritDoc
    */
   @Override
   public Stream<ChannelEvent> getEventStream(PulseIdRangeQuery query) {
      Stream<ChannelEvent> dummyEventStream =
            getDummyEventStream(query.getChannel(), query.getStartPulseId(), query.getEndPulseId())
                  .map(ce -> {
                     return (ChannelEvent) ce;
                  });
      return dummyEventStream;

   }

   /**
    * @{inheritDoc
    */
   @Override
   public Stream<ChannelEvent> getEventStream(TimeRangeQuery query) {
      Stream<ChannelEvent> dummyEventStream =
            getDummyEventStream(query.getChannel(), query.getStartMillis() / 10, query.getEndMillis() / 10)
                  .map(ce -> {
                     return (ChannelEvent) ce;
                  });
      return dummyEventStream;
   }


   private Stream<? extends DataEvent> getDummyEventStream(String channel, long startIndex, long endIndex) {
      String channelLower = channel.toLowerCase();

      Stream<? extends DataEvent> eventStream = LongStream.rangeClosed(startIndex, endIndex).mapToObj(i -> {
         if (channelLower.contains("waveform")) {
            long[] value = random.longs(2048).toArray();
            value[0] = i;
            return new ChannelEvent(
                  channel,
                  i * 10,
                  0,
                  i,
                  i * 10,
                  0,
                  KEYSPACE,
                  value
            );

         } else if (channelLower.contains("image")) {
            int x = 640;
            int y = 480;
            int[] shape = new int[] {x, y};
            long[] value = random.longs(x * y).toArray();
            value[0] = i;
            return new ChannelEvent(
                  channel,
                  i * 10,
                  0,
                  i,
                  i * 10,
                  0,
                  KEYSPACE,
                  value,
                  shape
            );
         } else {
            return new ChannelEvent(
                  channel,
                  i * 10,
                  0,
                  i,
                  i * 10,
                  0,
                  KEYSPACE,
                  i
            );
         }
      });

      return eventStream;
   }

   private List<? extends DataEvent> getDummyEvents(String channel, long startIndex, long endIndex) {
      return getDummyEventStream(channel, startIndex, endIndex).collect(Collectors.toList());
   }

   /**
    * @{inheritDoc
    */
   @Override
   public List<String> getChannels() {
      return Lists.newArrayList(channels);
   }

   /**
    * @{inheritDoc
    */
   @Override
   public List<String> getChannels(String regex) {
      return Lists.newArrayList(channels).stream().filter(s -> {
         return s.contains(regex);
      }).collect(Collectors.toList());
   }

   /**
    * @{inheritDoc
    */
   @Override
   public ChannelEvent getEvent(MetaChannelEvent queryInfo, String... columns) {
      if (queryInfo.getPulseId() > 0) {
         return (ChannelEvent) getDummyEvents(queryInfo.getChannel(), queryInfo.getPulseId(), queryInfo.getPulseId())
               .get(0);
      }
      return (ChannelEvent) getDummyEvents(queryInfo.getChannel(), queryInfo.getGlobalMillis() / 10,
            queryInfo.getGlobalMillis() / 10).get(0);
   }

   /**
    * @{inheritDoc
    */
   @Override
   public CompletableFuture<ChannelEvent> getEventAsync(MetaChannelEvent queryInfo, String... columns) {
      // implement when needed
      throw new UnsupportedOperationException();
   }

   /**
    * @{inheritDoc
    */
   @Override
   public Stream<? extends ChannelEventQuery> getChannelEventQueryStream(PulseIdRangeQuery query) {

      return dataGen.generateMetaPulseId(
            query.getStartPulseId(),
            (query.getEndPulseId() - query.getStartPulseId() + 1),
            i -> i * 10,
            i -> 0,
            i -> i,
            query.getChannel())
            .stream()
            .map(metaPulse -> {
               metaPulse.setKeyspace(KEYSPACE);
               return metaPulse;
            });
   }

   /**
    * @{inheritDoc
    */
   @Override
   public Stream<? extends ChannelEventQuery> getChannelEventQueryStream(TimeRangeQuery query) {

      return dataGen.generateMetaTime(
            KEYSPACE,
            query.getStartMillis() / 10,
            ((query.getEndMillis() - query.getStartMillis()) / 10 + 1),
            i -> i * 10,
            i -> 0,
            i -> i,
            query.getChannel()).stream();
   }

   /**
    * @{inheritDoc
    */
   @Override
   public Stream<MetaPulseId> getMetaStream(PulseIdRangeQuery query) {

      return getChannelEventQueryStream(query).map(r -> {
         return (MetaPulseId) r;
      });

   }


   /**
    * @{inheritDoc
    */
   @Override
   public Stream<? extends MetaChannelEvent> getMetaStream(TimeRangeQuery query) {

      return getChannelEventQueryStream(query).map(r -> {
         return (MetaChannelEvent) r;
      });
   }

   @Override
   public Stream<ChannelEvent> getEventStream(Stream<? extends MetaChannelEvent> queryInfos) {
      return getEventStream(null, queryInfos);
   }

   @Override
   public Stream<ChannelConfiguration> getChannelConfiguration(TimeRangeQuery query) {
      // implement when needed
      throw new UnsupportedOperationException();
   }

   @Override
   public ChannelConfiguration getChannelConfigurationBefore(TimeRangeQuery query) {
      try {
         return getChannelConfigurationBeforeAsync(query)
               .get(30, TimeUnit.SECONDS);
      } catch (Throwable t) {
         LOGGER.error("Could not read ChannelConfiguration from DB.", t);
         return null;
      }
   }

   @Override
   public CompletableFuture<ChannelConfiguration> getChannelConfigurationBeforeAsync(TimeRangeQuery query) {
      // implement when needed
      throw new UnsupportedOperationException();
   }

   @Override
   public ChannelConfiguration getChannelConfigurationAfter(TimeRangeQuery query) {
      try {
         return getChannelConfigurationAfterAsync(query)
               .get(30, TimeUnit.SECONDS);
      } catch (Throwable t) {
         LOGGER.error("Could not read ChannelConfiguration from DB.", t);
         return null;
      }
   }

   @Override
   public CompletableFuture<ChannelConfiguration> getChannelConfigurationAfterAsync(TimeRangeQuery query) {
      // implement when needed
      throw new UnsupportedOperationException();
   }
}
